from typing import List
from typeguard import typechecked
from dotenv import dotenv_values
from .env import Env


@typechecked
class EnvFile():

    def __init__(self, env_file: str, prefix: str = ''):
        self.env_file = env_file
        self.prefix = prefix.upper()

    def get_envs(self) -> List[Env]:
        envs: List[Env] = []
        env_map = dotenv_values(self.env_file)
        for key, value in env_map.items():
            os_name = key
            if self.prefix != '':
                os_name = f'{self.prefix}_{os_name}'
            envs.append(Env(name=key, os_name=os_name, default=value))
        return envs



import http.server
import socketserver
import os

MESSAGE = os.getenv('MESSAGE', 'Hello, world!')
PORT = int(os.getenv('PORT', '8000'))


class Handler(http.server.SimpleHTTPRequestHandler):
    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(MESSAGE.encode())


httpd = socketserver.TCPServer(("", PORT), Handler)

print(f"Serving at port {PORT}")
httpd.serve_forever()



import logging
import os
import pkg_resources


def get_version():
    try:
        dist = pkg_resources.get_distribution("zrb")
        return dist.version
    except pkg_resources.DistributionNotFound:
        import flit
        meta = flit.read_module_metadata("zrb")
        return meta["module"]["version"]


def get_logging_level():
    logging_level_str = os.getenv('ZRB_LOGGING_LEVEL', 'WARNING').lower()
    logging_level_map = {
        'critical': logging.CRITICAL,
        'error': logging.ERROR,
        'warning': logging.WARNING,
        'warn': logging.WARN,
        'info': logging.INFO,
        'debug': logging.DEBUG,
        'notset': logging.NOTSET,
    }
    if logging_level_str in logging_level_map:
        return logging_level_map[logging_level_str]
    return logging.WARNING


default_shell = os.getenv('ZRB_SHELL', 'bash')
init_scripts = os.getenv('ZRB_INIT_SCRIPTS', '').split(':')
logging_level = get_logging_level()
should_load_builtin = os.getenv('ZRB_SHOULD_LOAD_BUILTIN', '1') != '0'
env_prefix = os.getenv('ZRB_ENV', '')
show_advertisement = os.getenv('ZRB_SHOW_ADVERTISEMENT', '1') != '0'
version = get_version()



from zrb.builtin._group import project_group
from zrb import Task, runner

stop_project_containers = Task(
    name='stop-containers',
    group=project_group,
    upstreams=[],
    description='Stop project containers'
)
runner.register(stop_project_containers)



from typing import Any, List, Optional
from ...helper.middlewares.replacement import (
    Replacement, ReplacementMiddleware
)
from ...task.decorator import python_task
from ...task.task import Task
from ...task_input.str_input import StrInput
from ...task_input.int_input import IntInput
from ...helper.accessories.name import get_random_name
from ...helper.codemod.add_assert_resource import add_assert_resource
from ...helper.codemod.add_import_module import add_import_module
from ...helper import util
from ...helper.middlewares.replacement import (
    add_pascal_key, add_snake_key, add_camel_key,
    add_kebab_key, add_human_readable_key
)
import os


project_dir_input = StrInput(
    name='project-dir',
    shortcut='d',
    prompt='Project directory',
    default='.'
)

project_name_input = StrInput(
    name='project-name',
    shortcut='n',
    prompt='Project name (can be empty)',
    default=''
)

app_name_input = StrInput(
    name='app-name',
    shortcut='a',
    prompt='App name',
    default=get_random_name()
)

task_name_input = StrInput(
    name='task-name',
    shortcut='t',
    prompt='Task name',
    default=f'run-{get_random_name()}'
)
http_port_input = IntInput(
    name='http-port',
    shortcut='p',
    prompt='Port',
    default=8080
)

new_task_scaffold_lock = os.path.sep.join([
    '{{ os.path.join(input.project_dir, "_automate") }}',
    '{{ util.to_snake_case(input.task_name) }}.py'
])

new_app_scaffold_lock = os.path.sep.join([
    '{{ os.path.join(input.project_dir, "_automate") }}',
    '{{ util.to_snake_case(input.app_name) }}.py'
])


def validate_project_dir(project_dir: str):
    if not os.path.isfile(os.path.join(project_dir, 'zrb_init.py')):
        raise Exception(f'Not a project: {project_dir}')


def create_register_task_module(
    upstreams: Optional[List[Task]] = None
) -> Task:
    @python_task(
        name='register-task-module',
        inputs=[project_dir_input, task_name_input],
        upstreams=[] if upstreams is None else upstreams
    )
    def task(*args: Any, **kwargs: Any):
        task: Task = kwargs.get('_task')
        project_dir = kwargs.get('project_dir')
        validate_project_dir(project_dir)
        task_name = kwargs.get('task_name')
        task.print_out(f'Register module: _automate.{task_name}')
        register_module(
            project_dir=project_dir,
            module_name='.'.join([
                '_automate', util.to_snake_case(task_name)
            ])
        )
    return task


def create_register_local_app_module(
    upstreams: Optional[List[Task]] = None
) -> Task:
    return _create_register_app_module('local', upstreams)


def create_register_container_app_module(
    upstreams: Optional[List[Task]] = None
) -> Task:
    return _create_register_app_module('container', upstreams)


def create_register_deployment_app_module(
    upstreams: Optional[List[Task]] = None
) -> Task:
    return _create_register_app_module('deployment', upstreams)


def _create_register_app_module(
    module: str,
    upstreams: Optional[List[Task]] = None
) -> Task:
    @python_task(
        name='register-app-module',
        inputs=[project_dir_input, app_name_input],
        upstreams=[] if upstreams is None else upstreams
    )
    def task(*args: Any, **kwargs: Any):
        task: Task = kwargs.get('_task')
        project_dir = kwargs.get('project_dir')
        validate_project_dir(project_dir)
        app_name = kwargs.get('app_name')
        task.print_out(f'Register module: _automate.{app_name}.{module}')
        snake_app_name = util.to_snake_case(app_name)
        register_module(
            project_dir=project_dir,
            module_name='.'.join([
                '_automate', snake_app_name, module
            ]),
            alias=f'{snake_app_name}_{module}'
        )
    return task


def register_module(
    project_dir: str, module_name: str, alias: Optional[str] = None
):
    zrb_init_path = os.path.join(project_dir, 'zrb_init.py')
    with open(zrb_init_path, 'r') as f:
        code = f.read()
        if alias is None:
            alias = module_name.split('.')[-1]
        code = add_import_module(
            code=code,
            module_path=module_name,
            alias=alias
        )
        code = add_assert_resource(code, alias)
    with open(zrb_init_path, 'w') as f:
        f.write(code)
    return True


def get_default_task_replacements() -> Replacement:
    return {
        'taskName': '{{input.task_name}}',
        'appName': '{{input.app_name}}',
        'httpPort': '{{util.coalesce(input.http_port, "3000")}}',
    }


def get_default_task_replacement_middlewares() -> List[ReplacementMiddleware]:
    return [
        # task
        add_pascal_key('PascalTaskName', 'taskName'),
        add_camel_key('camelTaskName', 'taskName'),
        add_snake_key('snake_task_name', 'taskName'),
        add_kebab_key('kebab-task-name', 'taskName'),
        add_human_readable_key('human readable task name', 'taskName'),
        # app
        add_pascal_key('PascalAppName', 'appName'),
        add_camel_key('camelAppName', 'appName'),
        add_snake_key('snake_app_name', 'appName'),
        add_kebab_key('kebab-app-name', 'appName'),
        add_human_readable_key('human readable app name', 'appName'),
    ]




from typing import Any, Callable, List, Optional
from .._common import project_dir_input, app_name_input
from ....task.decorator import python_task
from ....task.task import Task
from ....helper import util
from ....helper.codemod.add_import_module import add_import_module
from ....helper.codemod.add_upstream_to_task import add_upstream_to_task
from .add import add_project_automation
from .task_factory_helper import (
    _get_app_local_task_file, _get_app_container_task_file,
    _get_app_deployment_task_file, _get_app_start_task_var,
    _get_app_start_container_task_var, _get_app_stop_container_task_var,
    _get_app_remove_container_var, _get_app_build_image_task_var,
    _get_app_push_image_task_var, _get_app_deploy_task_var,
    _get_app_remove_deployment_task_var
)
import os


def create_add_project_automation(
    upstreams: Optional[List[Task]] = None
) -> Task:
    @python_task(
        name='add-project-automation',
        inputs=[project_dir_input],
        upstreams=[] if upstreams is None else upstreams
    )
    def _task(*args: Any, **kwargs: Any):
        task: Task = kwargs.get('_task')
        project_dir = kwargs.get('project_dir', '.')
        task.print_out('Create project automation modules if not exist')
        add_project_automation(project_dir)
    return _task


def create_register_app_start(
    upstreams: Optional[List[Task]] = None,
):
    return _create_register_app_task(
        task_name='register-start',
        upstreams=upstreams,
        project_automation_file='start_project.py',
        project_automation_task_name='start',
        get_file_name=_get_app_local_task_file,
        get_task_var=_get_app_start_task_var,
    )


def create_register_app_start_container(
    upstreams: Optional[List[Task]] = None,
    as_local_task: bool = False
):
    project_automation_file = 'start_project_containers.py'
    if as_local_task:
        project_automation_file = 'start_project.py'
    project_automation_task_name = 'start-containers'
    if as_local_task:
        project_automation_task_name = 'start'
    return _create_register_app_task(
        task_name='register-start-container',
        upstreams=upstreams,
        project_automation_file=project_automation_file,
        project_automation_task_name=project_automation_task_name,
        get_file_name=_get_app_container_task_file,
        get_task_var=_get_app_start_container_task_var,
    )


def create_register_app_stop_container(
    upstreams: Optional[List[Task]] = None,
):
    return _create_register_app_task(
        task_name='register-stop-container',
        upstreams=upstreams,
        project_automation_file='stop_project_containers.py',
        project_automation_task_name='stop-containers',
        get_file_name=_get_app_container_task_file,
        get_task_var=_get_app_stop_container_task_var,
    )


def create_register_app_remove_container(
    upstreams: Optional[List[Task]] = None,
):
    return _create_register_app_task(
        task_name='register-remove-container',
        upstreams=upstreams,
        project_automation_file='remove_project_containers.py',
        project_automation_task_name='remove-containers',
        get_file_name=_get_app_container_task_file,
        get_task_var=_get_app_remove_container_var,
    )


def create_register_app_push_image(
    upstreams: Optional[List[Task]] = None,
):
    return _create_register_app_task(
        task_name='register-push-image',
        upstreams=upstreams,
        project_automation_file='push_project_images.py',
        project_automation_task_name='push-images',
        get_file_name=_get_app_container_task_file,
        get_task_var=_get_app_push_image_task_var
    )


def create_register_app_build_image(
    upstreams: Optional[List[Task]] = None,
):
    return _create_register_app_task(
        task_name='register-build-image',
        upstreams=upstreams,
        project_automation_file='build_project_images.py',
        project_automation_task_name='build-images',
        get_file_name=_get_app_container_task_file,
        get_task_var=_get_app_build_image_task_var
    )


def create_register_app_deploy(
    upstreams: Optional[List[Task]] = None,
):
    return _create_register_app_task(
        task_name='register-deploy',
        upstreams=upstreams,
        project_automation_file='deploy_project.py',
        project_automation_task_name='deploy',
        get_file_name=_get_app_deployment_task_file,
        get_task_var=_get_app_deploy_task_var
    )


def create_register_app_remove_deployment(
    upstreams: Optional[List[Task]] = None,
):
    return _create_register_app_task(
        task_name='register-remove-deployment',
        project_automation_file='remove_project_deployment.py',
        project_automation_task_name='remove-deployment',
        upstreams=upstreams,
        get_file_name=_get_app_deployment_task_file,
        get_task_var=_get_app_remove_deployment_task_var
    )


def _create_register_app_task(
    task_name: str,
    project_automation_file: str,
    project_automation_task_name: str,
    upstreams: Optional[List[Task]] = None,
    get_file_name: Callable[[str, str], str] = None,
    get_task_var: Callable[[str], str] = None,
):
    @python_task(
        name=task_name,
        inputs=[project_dir_input, app_name_input],
        upstreams=[] if upstreams is None else upstreams
    )
    def _task(*args: Any, **kwargs: Any):
        task: Task = kwargs.get('_task')
        project_dir = kwargs.get('project_dir', '.')
        app_name = kwargs.get('app_name')
        snake_app_name = util.to_snake_case(app_name)
        upstream_task_file = get_file_name(project_dir, snake_app_name)
        upstream_task_var = get_task_var(snake_app_name)
        project_automation_dir = os.path.join(
            project_dir, '_automate', '_project'
        )
        project_automation_path = os.path.join(
            project_automation_dir, f'{project_automation_file}'
        )
        upstream_task_rel_file_path = os.path.relpath(
            upstream_task_file, project_automation_path
        )
        # normalize `..` parts
        upstream_module_parts = [
            part if part != '..' else ''
            for part in upstream_task_rel_file_path.split(os.path.sep)
        ]
        # remove .py extenstion
        last_part = upstream_module_parts[-1]
        upstream_module_parts[-1] = os.path.splitext(last_part)[0]
        # turn into module path
        upstream_module_path = '.'.join(upstream_module_parts)
        task.print_out(f'Add {upstream_task_var} to project automation')
        with open(project_automation_path, 'r') as f:
            code = f.read()
            code = add_import_module(
                code=code,
                module_path=upstream_module_path,
                resource=upstream_task_var
            )
            code = add_upstream_to_task(
                code, project_automation_task_name, upstream_task_var
            )
        with open(project_automation_path, 'w') as f:
            f.write(code)

    return _task



from zrb.builtin._group import project_group
from zrb import Task, runner

build_project_images = Task(
    name='build-images',
    group=project_group,
    upstreams=[],
    description='Build project images'
)
runner.register(build_project_images)



from zrb import (
    DockerComposeTask, HTTPChecker, Env, runner
)
from zrb.builtin._group import project_group
import os

current_dir = os.path.dirname(__file__)
resource_dir = os.path.abspath(os.path.join(
    current_dir, '..', 'src', 'kebab-task-name'
))


snake_task_name = DockerComposeTask(
    name='kebab-task-name',
    description='human readable task name',
    group=project_group,
    cwd=resource_dir,
    compose_cmd='composeCommand',
    compose_env_prefix='ENV_PREFIX',
    envs=[
        Env(
            name='MESSAGE',
            os_name='ENV_PREFIX_MESSAGE',
            default='Salve Mane'
        ),
        Env(
            name='CONTAINER_PORT',
            os_name='ENV_PREFIX_CONTAINER_PORT',
            default='3000'
        ),
        Env(
            name='HOST_PORT',
            os_name='ENV_PREFIX_HOST_PORT',
            default='httpPort'
        ),
    ],
    checkers=[
        HTTPChecker(port='{{env.HOST_PORT}}')
    ]
)
runner.register(snake_task_name)



from zrb.helper.codemod.add_upstream_to_task import add_upstream_to_task


task_with_upstream_code = '''
a = Task(name='alpha', upstreams=[])
b = Task(name='beta', upstreams=[a])
c = Task(name='gamma', upstreams=[b])
'''

new_task_with_upstream_code = '''
a = Task(name='alpha', upstreams=[])
b = Task(name='beta', upstreams=[a, delta])
c = Task(name='gamma', upstreams=[b])
'''


def test_add_upstream_to_task_with_upstream():
    code = add_upstream_to_task(
        task_with_upstream_code, 'beta', 'delta'
    )
    assert code == new_task_with_upstream_code


task_without_upstream_code = '''
a = Task(name='alpha')
b = Task(name='beta')
c = Task(name='gamma')
'''

new_task_without_upstream_code = '''
a = Task(name='alpha')
b = Task(name='beta', upstreams = [delta])
c = Task(name='gamma')
'''


def test_add_upstream_to_task_without_upstream():
    code = add_upstream_to_task(
        task_without_upstream_code, 'beta', 'delta'
    )
    assert code == new_task_without_upstream_code



from ....helper.file.copy_tree import copy_tree
from ....helper.codemod.add_assert_resource import add_assert_resource
from ....helper.codemod.add_import_module import add_import_module

import os

current_dir = os.path.dirname(__file__)


def add_project_automation(project_dir: str):
    if os.path.exists(os.path.join(project_dir, '_automate', '_project')):
        return
    copy_tree(
        src=os.path.join(current_dir, 'template'),
        dst=project_dir
    )
    project_task_module_path = '_automate._project'
    zrb_init_path = os.path.join(project_dir, 'zrb_init.py')
    with open(zrb_init_path, 'r') as f:
        code = f.read()
        import_alias = project_task_module_path.split('.')[-1]
        code = add_import_module(
            code=code,
            module_path=project_task_module_path,
            alias=import_alias
        )
        code = add_assert_resource(code, import_alias)
    with open(zrb_init_path, 'w') as f:
        f.write(code)



from zrb.builtin._group import project_group
from zrb import Task, runner

remove_project_containers = Task(
    name='remove-containers',
    group=project_group,
    upstreams=[],
    description='Remove project containers'
)
runner.register(remove_project_containers)



from typing import Any
from ..._group import project_add_group
from ....task.task import Task
from ....task.decorator import python_task
from ....task_input.str_input import StrInput
from ....task.resource_maker import ResourceMaker
from ....runner import runner
from .._common import (
    project_dir_input, task_name_input, http_port_input, validate_project_dir,
    create_register_task_module, get_default_task_replacements,
    get_default_task_replacement_middlewares, new_task_scaffold_lock
)
from ....helper import util

import os

# Common definitions

current_dir = os.path.dirname(__file__)


@python_task(
    name='task-validate-create',
    inputs=[project_dir_input, task_name_input],
)
def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_project_dir(project_dir)
    task_name = kwargs.get(project_dir, 'task_name')
    automation_file = os.path.join(
        project_dir, '_automate', f'{util.to_snake_case(task_name)}.py'
    )
    if os.path.isfile(automation_file):
        raise Exception(f'File already exists: {automation_file}')


replacements = get_default_task_replacements()
replacements.update({
    'composeCommand': '{{ util.coalesce(input.compose_command, "up") }}',
    'ENV_PREFIX': '{{ util.coalesce(input.env_prefix, "MY").upper() }}'
})
copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        task_name_input,
        http_port_input,
        StrInput(
            name='compose-command', prompt='Compose command', default='up'
        ),
        StrInput(
            name='env-prefix', prompt='Env prefix', default='MY'
        ),
    ],
    upstreams=[validate],
    replacements=replacements,
    replacement_middlewares=get_default_task_replacement_middlewares(),
    template_path=os.path.join(current_dir, 'template'),
    destination_path='{{ input.project_dir }}',
    scaffold_locks=[new_task_scaffold_lock]
)

register_task_module = create_register_task_module(
    upstreams=[copy_resource]
)


@python_task(
    name='docker-compose-task',
    group=project_add_group,
    upstreams=[register_task_module],
    runner=runner
)
def add_docker_compose_task(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    task.print_out('Success')



from typing import Any, Iterable, Mapping, Optional, Union
from typeguard import typechecked
from ..helper.accessories.color import (
    get_random_color, is_valid_color, colored
)
from ..helper.accessories.icon import get_random_icon
from ..helper.string.conversion import (
    to_cmd_name, to_variable_name
)
from ..helper.render_data import DEFAULT_RENDER_DATA
from ..helper.log import logger
from ..task_env.env import Env
from ..task_env.env_file import EnvFile
from ..task_group.group import Group
from ..task_input._constant import RESERVED_INPUT_NAMES

import datetime
import os
import sys
import time
import jinja2


MAX_NAME_LENGTH = 20
MULTILINE_INDENT = ' ' * 8


class AnyExtensionFileSystemLoader(jinja2.FileSystemLoader):
    def get_source(self, environment, template):
        for search_dir in self.searchpath:
            file_path = os.path.join(search_dir, template)
            if os.path.exists(file_path):
                with open(file_path, 'r') as file:
                    contents = file.read()
                    return contents, file_path, lambda: False
        raise jinja2.TemplateNotFound(template)


@typechecked
class TimeTracker():

    def __init__(self):
        self._start_time: float = 0
        self._end_time: float = 0

    def start_timer(self):
        self._start_time = time.time()

    def end_timer(self):
        self._end_time = time.time()

    def get_elapsed_time(self) -> float:
        return self._end_time - self._start_time


@typechecked
class AttemptTracker():

    def __init__(self, retry: int = 2):
        self.retry = retry
        self._attempt: int = 1

    def get_max_attempt(self) -> int:
        return self.retry + 1

    def get_attempt(self) -> int:
        return self._attempt

    def increase_attempt(self):
        self._attempt += 1

    def should_attempt(self) -> bool:
        attempt = self.get_attempt()
        max_attempt = self.get_max_attempt()
        return attempt <= max_attempt

    def is_last_attempt(self) -> bool:
        attempt = self.get_attempt()
        max_attempt = self.get_max_attempt()
        return attempt >= max_attempt


@typechecked
class FinishTracker():

    def __init__(self):
        self._is_done: bool = False

    def mark_as_done(self):
        self._is_done = True

    def mark_as_undone(self):
        self._is_done = False

    def is_done(self) -> bool:
        return self._is_done


@typechecked
class PidModel():

    def __init__(self):
        self.zrb_task_pid: int = os.getpid()

    def set_task_pid(self, pid: int):
        self.zrb_task_pid = pid

    def get_task_pid(self) -> int:
        return self.zrb_task_pid


@typechecked
class TaskDataModel():

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
    ):
        self.name = name
        self.group = group
        self.envs = envs
        self.env_files = env_files
        for key in dict(os.environ):
            self.envs.append(Env(name=key, os_name=key))
        self.icon = icon
        self.color = color
        self._input_map: Mapping[str, Any] = {}
        self._env_map: Mapping[str, str] = {}
        self._complete_name: Optional[str] = None
        self._is_keyval_set = False  # Flag
        self._has_cli_interface = False
        self._render_data: Optional[Mapping[str, Any]] = None

    def get_icon(self) -> str:
        if self.icon is None or self.icon == '':
            self.icon = get_random_icon()
        return self.icon

    def get_color(self) -> str:
        if self.color is None or not is_valid_color(self.color):
            self.color = get_random_color()
        return self.color

    def get_cmd_name(self) -> str:
        return to_cmd_name(self.name)

    def log_debug(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.debug(colored_message)

    def log_warn(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.warning(colored_message)

    def log_info(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.info(colored_message)

    def log_error(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', color='red', attrs=['bold']
        )
        logger.error(colored_message, exc_info=True)

    def log_critical(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', color='red', attrs=['bold']
        )
        logger.critical(colored_message, exc_info=True)

    def print_out(self, msg: Any):
        prefix = self._get_colored_print_prefix()
        print(f'🤖 ➜  {prefix} • {msg}'.rstrip(), file=sys.stderr)

    def print_err(self, msg: Any):
        prefix = self._get_colored_print_prefix()
        print(f'🤖 ⚠  {prefix} • {msg}'.rstrip(), file=sys.stderr)

    def print_out_dark(self, msg: Any):
        self.print_out(colored(msg, attrs=['dark']))

    def colored(self, text: str) -> str:
        return colored(text, color=self.get_color())

    def play_bell(self):
        print('\a', end='', file=sys.stderr)

    def _get_colored_print_prefix(self) -> str:
        return self.colored(self._get_print_prefix())

    def _get_print_prefix(self) -> str:
        common_prefix = self._get_common_prefix(show_time=True)
        icon = self.get_icon()
        truncated_name = self._get_truncated_complete_name()
        return f'{common_prefix} • {icon} {truncated_name}'

    def _get_log_prefix(self) -> str:
        common_prefix = self._get_common_prefix(show_time=False)
        icon = self.get_icon()
        filled_name = self._get_filled_complete_name()
        return f'{common_prefix} • {icon} {filled_name}'

    def _get_common_prefix(self, show_time: bool) -> str:
        attempt = self.get_attempt()
        max_attempt = self.get_max_attempt()
        pid = self.get_task_pid()
        if show_time:
            now = datetime.datetime.now().isoformat()
            return f'{now} ⚙ {pid} ➤ {attempt} of {max_attempt}'
        return f'⚙ {pid} ➤ {attempt} of {max_attempt}'

    def _get_filled_complete_name(self) -> str:
        return self._get_complete_name().rjust(MAX_NAME_LENGTH, ' ')

    def _get_truncated_complete_name(self):
        filled_name = self._get_filled_complete_name()
        if len(filled_name) > MAX_NAME_LENGTH:
            return '..' + filled_name[-(MAX_NAME_LENGTH-2):]
        return filled_name

    def get_input_map(self) -> Mapping[str, Any]:
        return self._input_map

    def get_env_map(self) -> Mapping[str, str]:
        return self._env_map

    def _inject_env_map(
        self, env_map: Mapping[str, str], override: bool = False
    ):
        for key, val in env_map.items():
            if override or key not in self._env_map:
                self._env_map[key] = val

    def render_float(self, val: Union[str, float]) -> float:
        if isinstance(val, str):
            return float(self.render_str(val))
        return val

    def render_int(self, val: Union[str, int]) -> int:
        if isinstance(val, str):
            return int(self.render_str(val))
        return val

    def render_bool(self, val: Union[str, bool]) -> bool:
        if isinstance(val, str):
            str_val = self.render_str(val)
            if str_val.lower() in ['true', '1', 'yes', 'y', 'active']:
                return True
            if str_val.lower() in ['false', '0', 'no', 'n', 'inactive']:
                return False
            raise Exception(f'Cannot infer boolean value from {str_val}')
        return val

    def render_str(self, val: str) -> str:
        template = jinja2.Template(val)
        data = self._get_render_data()
        self.log_debug(f'Render string template: {val}, with data: {data}')
        rendered_text = template.render(data)
        self.log_debug(f'Get rendered result: {rendered_text}')
        return rendered_text

    def render_file(self, location: str) -> str:
        location_dir = os.path.dirname(location)
        env = jinja2.Environment(
            loader=AnyExtensionFileSystemLoader([location_dir])
        )
        template = env.get_template(location)
        data = self._get_render_data()
        data['TEMPLATE_DIR'] = location_dir
        self.log_debug(f'Render template file: {template}, with data: {data}')
        rendered_text = template.render(data)
        self.log_debug(f'Get rendered result: {rendered_text}')
        return rendered_text

    def _get_render_data(self) -> Mapping[str, Any]:
        if self._render_data is not None:
            return self._render_data
        render_data = dict(DEFAULT_RENDER_DATA)
        render_data.update({
            'env': self._env_map,
            'input': self._input_map,
        })
        self._render_data = render_data
        return render_data

    def _get_multiline_repr(self, text: str) -> str:
        lines_repr: Iterable[str] = []
        for index, line in enumerate(text.split('\n')):
            line_number_repr = str(index + 1).rjust(4, '0')
            lines_repr.append(f'{MULTILINE_INDENT}{line_number_repr} | {line}')
        return '\n'.join(lines_repr)

    def _set_local_keyval(
        self,
        input_map: Mapping[str, Any],
        env_prefix: str = ''
    ):
        if self._is_keyval_set:
            return True
        self._is_keyval_set = True
        # Add self.inputs to input_map
        self.log_info('Set input map')
        for input_name, val in input_map.items():
            self._input_map[self._get_normalized_input_key(input_name)] = val
        self.log_debug(f'Input map: {self._input_map}')
        # Construct envs based on self.env_files and self.envs
        self.log_info('Merging env_files and envs')
        envs: Iterable[Env] = []
        for env_file in self.env_files:
            envs = envs + env_file.get_envs()
        envs = envs + self.envs
        # Add envs to env_map
        self.log_info('Set env map')
        for task_env in envs:
            env_name = task_env.name
            env_value = task_env.get(env_prefix)
            if self._is_probably_jinja(env_value):
                env_value = self.render_str(env_value)
            self._env_map[env_name] = env_value
        self.log_debug(f'Env map: {self._env_map}')

    def _is_probably_jinja(self, string: str) -> bool:
        if '{{' in string and '}}' in string:
            return True
        if '{%' in string and '%}' in string:
            return True
        return False

    def _get_normalized_input_key(self, key: str) -> str:
        if key in RESERVED_INPUT_NAMES:
            return key
        return to_variable_name(key)

    def _get_complete_name(self) -> str:
        if self._complete_name is not None:
            return self._complete_name
        executable_prefix = ''
        if self._has_cli_interface:
            executable_prefix += self._get_executable_name() + ' '
        cmd_name = self.get_cmd_name()
        if self.group is None:
            self._complete_name = f'{executable_prefix}{cmd_name}'
            return self._complete_name
        group_cmd_name = self.group.get_complete_name()
        self._complete_name = f'{executable_prefix}{group_cmd_name} {cmd_name}'
        return self._complete_name

    def _get_executable_name(self) -> str:
        if len(sys.argv) > 0 and sys.argv[0] != '':
            return os.path.basename(sys.argv[0])
        return 'zrb'

    def set_has_cli_interface(self):
        self._has_cli_interface = True


@typechecked
class TaskModel(
    TaskDataModel, PidModel, FinishTracker, AttemptTracker, TimeTracker
):

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        retry: int = 2,
    ):
        TaskDataModel.__init__(
            self,
            name=name,
            group=group,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color
        )
        retry = self.ensure_non_negative(retry, 'Find negative retry')
        PidModel.__init__(self)
        FinishTracker.__init__(self)
        AttemptTracker.__init__(self, retry=retry)
        TimeTracker.__init__(self)

    def ensure_non_negative(self, value: float, error_label: str) -> float:
        if value < 0:
            self.log_warn(
                f'{error_label}: {value}'
            )
            return 0
        return value

    def show_celebration(self):
        complete_name = self._get_complete_name()
        elapsed_time = self.get_elapsed_time()
        message = '\n'.join([
            f'{complete_name} completed in {elapsed_time} seconds',
        ])
        print(colored(message, attrs=['dark']), file=sys.stderr)
        self.play_bell()



from typing import Any
from ..._group import project_add_group
from ....task.task import Task
from ....task.decorator import python_task
from ....task_input.str_input import StrInput
from ....task.resource_maker import ResourceMaker
from ....runner import runner
from .._common import (
    project_dir_input, app_name_input, http_port_input, validate_project_dir,
    create_register_local_app_module, create_register_container_app_module,
    create_register_deployment_app_module, get_default_task_replacements,
    get_default_task_replacement_middlewares, new_app_scaffold_lock
)
from ..project_task.task_factory import (
    create_add_project_automation, create_register_app_start,
    create_register_app_start_container, create_register_app_stop_container,
    create_register_app_remove_container, create_register_app_build_image,
    create_register_app_push_image,
)
from ....helper import util

import os

# Common definitions

current_dir = os.path.dirname(__file__)


@python_task(
    name='task-validate-create',
    inputs=[project_dir_input, app_name_input],
)
def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_project_dir(project_dir)
    app_name = kwargs.get('app_name')
    automation_dir = os.path.join(
        project_dir, '_automate', util.to_snake_case(app_name)
    )
    if os.path.isdir(automation_dir):
        raise Exception(f'Directory already exists: {automation_dir}')


replacements = get_default_task_replacements()
replacements.update({
    'ENV_PREFIX': '{{ util.coalesce(input.env_prefix, "MY").upper() }}'
})
copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        app_name_input,
        http_port_input,
        StrInput(
            name='env-prefix', prompt='Env prefix', default='MY'
        ),
    ],
    upstreams=[validate],
    replacements=replacements,
    replacement_middlewares=get_default_task_replacement_middlewares(),
    template_path=os.path.join(current_dir, 'template'),
    destination_path='{{ input.project_dir }}',
    scaffold_locks=[new_app_scaffold_lock]
)

register_local_app_module = create_register_local_app_module(
    upstreams=[copy_resource]
)

register_container_app_module = create_register_container_app_module(
    upstreams=[copy_resource]
)

register_deployment_app_module = create_register_deployment_app_module(
    upstreams=[copy_resource]
)

add_project_task = create_add_project_automation(
    upstreams=[copy_resource]
)

register_start = create_register_app_start(
    upstreams=[add_project_task]
)

register_start_container = create_register_app_start_container(
    upstreams=[add_project_task]
)

register_stop_container = create_register_app_stop_container(
    upstreams=[add_project_task]
)

register_remove_container = create_register_app_remove_container(
    upstreams=[add_project_task]
)

register_push_image = create_register_app_push_image(
    upstreams=[add_project_task]
)

register_build_image = create_register_app_build_image(
    upstreams=[add_project_task]
)


@python_task(
    name='simple-python-app',
    group=project_add_group,
    inputs=[project_dir_input, app_name_input],
    upstreams=[
        register_local_app_module,
        register_container_app_module,
        register_deployment_app_module,
        register_start,
        register_start_container,
        register_stop_container,
        register_remove_container,
        register_build_image,
        register_push_image
    ],
    runner=runner
)
def add_simple_python_app(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    task.print_out('Success')



import os


def _get_app_local_task_file(project_dir: str, snake_app_name: str) -> str:
    return os.path.join(
        project_dir, '_automate', snake_app_name, 'local.py'
    )


def _get_app_container_task_file(project_dir: str, snake_app_name: str) -> str:
    return os.path.join(
        project_dir, '_automate', snake_app_name, 'container.py'
    )


def _get_app_deployment_task_file(
    project_dir: str, snake_app_name: str
) -> str:
    return os.path.join(
        project_dir, '_automate', snake_app_name, 'deployment.py'
    )


def _get_app_start_task_var(snake_app_name: str) -> str:
    return f'start_{snake_app_name}'


def _get_app_start_container_task_var(snake_app_name: str) -> str:
    return f'start_{snake_app_name}_container'


def _get_app_stop_container_task_var(snake_app_name: str) -> str:
    return f'stop_{snake_app_name}_container'


def _get_app_remove_container_var(snake_app_name: str) -> str:
    return f'remove_{snake_app_name}_container'


def _get_app_push_image_task_var(snake_app_name: str) -> str:
    return f'push_{snake_app_name}_image'


def _get_app_build_image_task_var(snake_app_name: str) -> str:
    return f'build_{snake_app_name}_image'


def _get_app_deploy_task_var(snake_app_name: str) -> str:
    return f'deploy_{snake_app_name}_image'


def _get_app_remove_deployment_task_var(snake_app_name: str) -> str:
    return f'remove_{snake_app_name}_deployment'
